#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count
from django.utils.functional import cached_property


class PotQuerySet(models.QuerySet):
    def _count_relation(self, field, annotation_name):
        return self.annotate(**{
            annotation_name: Count(field, distinct=True)
        })

    def with_tea_count(self):
        return self._count_relation('supported_teas', 'tea_count')

    def with_addition_count(self):
        return self._count_relation('supported_additions', 'addition_count')


class Pot(models.Model):
    """A Tea- or Coffee Pot capable of brewing a choice beverage."""
    name = models.CharField(
        max_length=35,
        unique=True,
        help_text='The name of this pot, e.g. "Joe\'s Joe Jar" or "Breville (R)'
                  ' BTM800XL"',
    )

    brew_coffee = models.BooleanField(
        verbose_name='able to brew coffee',
        default=True,
        help_text="Can this pot brew coffee?",
    )

    supported_teas = models.ManyToManyField(
        'TeaType',
        blank=True,
        related_name='pot_list'
    )

    supported_additions = models.ManyToManyField(
        'Addition',
        blank=True,
        related_name='pot_list'
    )

    objects = PotQuerySet.as_manager()

    def __str__(self):
        return '{} - {}'.format(self.id, self.name)

    @cached_property
    def tea_capable(self):
        """Return True if this pot can serve tea."""
        return self.supported_teas.exists()

    @property
    def is_teapot(self):
        """Return True if this pot can serve tea, but cannot serve coffee."""
        return self.tea_capable and not self.brew_coffee

    def fetch_additions(self, addition_names):
        """
        Return the Additions that this pot supports whose names are in the
        provided sequence.

        If this pot does not support an Addition whose name is provided, raise
        an Addition.DoesNotExist error.
        """
        additions = self.supported_additions.filter(name__in=addition_names)
        # Resolve the query set to compute its length.
        # Replace this len() call with a queryset annotation if further
        # filtering becomes necessary.
        if len(addition_names) != len(additions):
            raise Addition.DoesNotExist
        return additions


class TeaType(models.Model):
    """
    A variety of tea that can be brewed by a pot.

    Per the HTCPCP standard, tea may be available as tea bags or tea leaves.
    """
    name = models.CharField(
        verbose_name='Tea name',
        max_length=35,
        unique=True,
        db_index=True,
    )

    slug = models.SlugField(unique=True)

    def __str__(self):
        return self.name


class Addition(models.Model):
    """
    A beverage addition that may be specified in the Accept-Additions header
    field of an HTCPCP request.
    """

    MILK = 'MLK'

    SYRUP = 'SYP'

    SWEETENER = 'SWT'

    SPICE = 'SPC'

    ALCOHOL = 'ACL'

    SUGAR = 'SUG'

    OTHER = 'OTR'

    TYPE_CHOICES = (
        (MILK, "Milk"),
        (SYRUP, "Syrup"),
        (SWEETENER, "Sweetener"),
        (SPICE, "Spice"),
        (ALCOHOL, "Alcohol"),
        (SUGAR, "Sugar"),
        (OTHER, "Other"),
    )

    name = models.CharField(
        max_length=35,
        unique=True,
        help_text="The name of this beverage addition as it would appear in the"
                  " HTCPCP Accept-Additions header field.",
    )

    type = models.CharField(
        max_length=3,
        choices=TYPE_CHOICES,
        verbose_name='Addition type',
    )

    def clean(self):
        # Calling parent in case it is given a non-empty body in the
        # future (currently it is empty).
        super().clean()

        error_msg = ("Decaffeinated addition not allowed. RFC 2324 specifies "
                     "no option for decaffeinated coffee. What's the point? ")

        if 'decaffeinated' in self.name.lower():
            raise ValidationError(error_msg)

    def __str__(self):
        return "{} / {}".format(self.get_type_display(), self.name)

    @property
    def is_milk(self):
        return self.type == self.MILK



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.urls import path

from .views import brew_pot

urlpatterns = [
    path('', brew_pot, name='htcpcp-index'),
    path('pot-<int:pot_designator>/', brew_pot, name='pot-detail'),
    path('pot-<int:pot_designator>/<str:tea_type>/', brew_pot, name='pot-detail-tea')

]



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.apps import AppConfig


class HTCPCPTeaConfig(AppConfig):
    name = 'django_htcpcp_tea'
    verbose_name = 'HTCPCP-TEA Server'



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.conf import settings as django_settings


class _HTCPCPTeaSettings:
    """
    Proxy to the standard Django settings that provides defaults for some
    of this app's settings.
    """

    ALLOW_DEPRECATED_POST = True

    DISABLE_CSRF = True

    GET_ADDITIONS = True

    POT_SESSIONS = True

    STRICT_MIME_TYPE = True

    STRICT_REQUEST_BODY = False

    def __init__(self, settings_prefix):
        self.prefix = settings_prefix

    def __getattribute__(self, item):
        try:
            return getattr(django_settings, '{}_{}'.format(
                object.__getattribute__(self, 'prefix'),
                item,
            ))
        except AttributeError:
            return object.__getattribute__(self, item)


htcpcp_settings = _HTCPCPTeaSettings('HTCPCP')



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.urls import reverse

from .models import Pot
from .settings import htcpcp_settings


def build_alternates(index_pot=None):
    """
    Generate the Alternates pairs for available beverages, optionally
    for a specific pot.
    """
    if index_pot:
        pots = (index_pot,)
    else:
        pots = Pot.objects.prefetch_related('supported_teas').all()
    for pot in pots:
        if pot.brew_coffee:
            yield (reverse('pot-detail', args=[pot.id]), 'message/coffeepot')
        for tea in pot.supported_teas.all():
            yield (reverse('pot-detail-tea', args=[pot.id, tea.slug]), 'message/teapot')


def resolve_requested_additions(request):
    """
    Return to requested additions for the provided request.

    Additions may be requested in the ``Accept-Additions`` header field, or
    (if the ``HTCPCP_GET_ADDITIONS`` settings is enabled) in the query string
    of a uri.

    Note that the returned additions are not guaranteed to be valid additions
    that are supported by any pot.
    """
    try:
        header = request.META['HTTP_ACCEPT_ADDITIONS']
        additions = [addition.strip() for addition in header.split(',')]
    except KeyError:
        additions = []

    if htcpcp_settings.GET_ADDITIONS:
        additions += request.GET.dict().keys()

    return additions



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from .settings import htcpcp_settings


class HTCPCPTeaMiddleware:

    HTCPCP_MESSAGE_KEYWORDS = (b'start', b'stop')

    def __init__(self, get_response):
        self.get_response = get_response
        self.valid_methods = ('BREW', 'WHEN')
        if htcpcp_settings.ALLOW_DEPRECATED_POST:
            self.valid_methods += ('POST',)

    def __call__(self, request):
        htcpcp_valid = True

        if request.method not in self.valid_methods:
            htcpcp_valid = False

        # Resolve HTCPCP message type (start or stop)
        if htcpcp_settings.STRICT_REQUEST_BODY:
            if request.body not in self.HTCPCP_MESSAGE_KEYWORDS:
                htcpcp_valid = False
            else:
                request.htcpcp_message_type = request.body.decode(encoding='utf-8')
        else:
            for keyword in self.HTCPCP_MESSAGE_KEYWORDS:
                if keyword in request.body:
                    request.htcpcp_message_type = keyword.decode(encoding='utf-8')
                    break  # Trigger else branch if no keyword is found
            else:
                htcpcp_valid = False

        request.htcpcp_valid = htcpcp_valid

        response = self.get_response(request)

        try:
            response['Alternates'] = self.build_alternates(response)
        except AttributeError:
            pass

        return response

    def build_alternates(self, response):
        fmt = '{{"{}" {{type {}}}}}'
        return ','.join(
            fmt.format(*pair) for pair in response.htcpcp_alternates
        )



# Generated by Django 2.2.2 on 2019-06-19 10:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_htcpcp_tea', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='addition',
            name='name',
            field=models.CharField(help_text='The name of this beverage addition as it would appear in the HTCPCP Accept-Additions header field.', max_length=35, unique=True),
        ),
        migrations.AlterField(
            model_name='pot',
            name='brew_coffee',
            field=models.BooleanField(default=True, help_text='Can this pot brew coffee?', verbose_name='able to brew coffee'),
        ),
        migrations.AlterField(
            model_name='pot',
            name='name',
            field=models.CharField(help_text='The name of this pot, e.g. "Joe\'s Joe Jar" or "Breville (R) BTM800XL"', max_length=35, unique=True),
        ),
        migrations.AlterField(
            model_name='pot',
            name='supported_additions',
            field=models.ManyToManyField(blank=True, related_name='pot_list', to='django_htcpcp_tea.Addition'),
        ),
        migrations.AlterField(
            model_name='pot',
            name='supported_teas',
            field=models.ManyToManyField(blank=True, related_name='pot_list', to='django_htcpcp_tea.TeaType'),
        ),
    ]



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Count

from .models import Addition, Pot, TeaType


class RelatedItemsExistsListFilter(admin.SimpleListFilter):
    """Admin list filter for whether an object has a ManyToMany related item."""
    related_item_field = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.related_item_field:
            error_msg = 'The list filter {} does not specify a related_item_field'
            raise ImproperlyConfigured(error_msg.format(self.__class__.__name__))

    def lookups(self, request, model_admin):
        return (
            ('y', 'Yes'),
            ('n', 'No'),
        )

    def queryset(self, request, queryset):
        value = self.value()
        lookup_param = '{}__isnull'.format(self.related_item_field)

        if value == 'y':
            return queryset.filter(**{lookup_param: False})

        if value == 'n':
            return queryset.filter(**{lookup_param: True})


class BrewTeaListFilter(RelatedItemsExistsListFilter):
    """Admin list filter for whether a pots serves any teas."""
    title = 'able to brew tea'

    parameter_name = 'brew_tea'

    related_item_field = 'supported_teas'


class ServedByAPotListFilter(RelatedItemsExistsListFilter):
    """Admin list filter for whether an object is served by a pot."""
    title = 'served by a pot'

    parameter_name = 'is_served'

    related_item_field = 'pot_list'


@admin.register(Pot)
class PotAdmin(admin.ModelAdmin):
    search_fields = ('supported_teas__name', 'supported_additions__name')

    fields = (('name', 'brew_coffee'), 'supported_teas', 'supported_additions')

    list_display = ('id', 'name', 'brew_coffee', 'tea_capable_view', 'tea_count_view', 'addition_count_view')

    list_display_links = ('id', 'name')

    filter_horizontal = ('supported_teas', 'supported_additions')

    list_filter = ('brew_coffee', BrewTeaListFilter, ('supported_teas', admin.RelatedOnlyFieldListFilter))

    save_as = True

    def tea_capable_view(self, obj):
        """Display whether the given pot can brew tea."""
        return obj.tea_capable

    tea_capable_view.boolean = True
    tea_capable_view.short_description = "able to brew tea"
    tea_capable_view.admin_order_field = 'supported_teas'

    def tea_count_view(self, obj):
        "Display the number of tea types that the given pot supports."
        return obj.tea_count

    tea_count_view.admin_order_field = 'tea_count'
    tea_count_view.short_description = 'supported teas'

    def addition_count_view(self, obj):
        "Display the number of addition types that the given pot supports."
        return obj.addition_count

    addition_count_view.admin_order_field = 'addition_count'
    addition_count_view.short_description = 'supported additions'

    def get_queryset(self, request):
        return super().get_queryset(request).with_tea_count().with_addition_count()


class PotsServingMixin:
    """Mixin to add a 'pots serving' item of a model admin's list_display."""

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        serving_callable = self.pots_serving_count_view.__name__
        if serving_callable not in fields:
            return fields + (serving_callable,)
        return fields

    def get_list_filter(self, request):
        return (ServedByAPotListFilter,) + super().get_list_filter(request)

    def pots_serving_count_view(self, obj):
        """Display the number of pots that serve the given object."""
        return obj.pot_count

    pots_serving_count_view.admin_order_field = 'pot_count'
    pots_serving_count_view.short_description = 'pots serving'

    def get_queryset(self, request):
        return super().get_queryset(request).annotate(pot_count=Count('pot_list'))


@admin.register(TeaType)
class TeaTypeAdmin(PotsServingMixin, admin.ModelAdmin):
    search_fields = ('name',)

    prepopulated_fields = {'slug': ('name',)}

    fieldsets = (
        (None, {
            'fields': ('name',)
        }),
        ('Advanced', {
            'classes': ('collapse',),
            'fields': ('slug',),
        }),
    )


@admin.register(Addition)
class AdditionAdmin(PotsServingMixin, admin.ModelAdmin):
    search_fields = ('name',)

    list_display = ('name', 'type')

    list_filter = ('type',)

    radio_fields = {'type': admin.HORIZONTAL}



#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

from datetime import datetime
from functools import wraps

from django.http import Http404
from django.shortcuts import get_object_or_404, render

from .models import Pot, Addition
from .settings import htcpcp_settings
from .utils import build_alternates, resolve_requested_additions


def require_htcpcp(func):
    """Decorator to make a view only respond to valid HTCPCP requests."""

    @wraps(func)
    def _require_htcpcp(request, *args, **kwargs):
        if not request.htcpcp_valid:
            raise Http404
        return func(request, *args, **kwargs)

    return _require_htcpcp


@require_htcpcp
def brew_pot(request, pot_designator=None, tea_type=None):
    if not pot_designator:
        response = render(request, 'django_htcpcp_tea/options.html', status=300)
        response.htcpcp_alternates = build_alternates()
        return response

    if request.method == 'WHEN' and request.htcpcp_message_type == 'start':
        return render(
            request,
            'django_htcpcp_tea/400.html',
            {'error_reason': 'Cannot start a beverage with a WHEN request.'},
            status=400
        )

    pot = get_object_or_404(Pot, id=pot_designator)

    if _request_for_tea(request, tea_type):
        response = _precheck_teapot(request, pot, tea_type)
        # Beverage name only required when starting a new beverage
        beverage_name = '{} Tea'.format(tea_type.capitalize) if tea_type else None
    else:
        response = _precheck_coffee(request, pot)
        beverage_name = 'coffee'

    if response is None:
        addition_names = resolve_requested_additions(request)
        try:
            additions = list(pot.fetch_additions(addition_names))
        except Addition.DoesNotExist:
            # Fetch all supported additions for the requested pot.
            # Note that this will result in an additional query.
            context = {'supported_additions': pot.supported_additions.all()}
            return render(request, 'django_htcpcp_tea/406.html', context, status=406)

        if htcpcp_settings.POT_SESSIONS:
            response = _finalize_beverage_with_session(request, pot, beverage_name, additions)
        else:
            response = _finalize_beverage(request, pot, beverage_name, additions)

    return response


def _request_for_tea(request, tea_type):
    """
    Determine whether the given request is for tea.

    If the ``STRICT_MIME_TYPE`` settings is enabled, then the MIME type of the
    request is used to unambiguously categorize it as a tea or coffee request.

    If strict MIME types are not enforced, the requested beverage type is
    inferred from the existence of a tea type in the request URI and the HTCPCP
    MIME type, if one is provided.
    """
    if request.content_type == 'message/teapot':
        return True
    elif not htcpcp_settings.STRICT_MIME_TYPE:
        return tea_type
    else:
        return False


def _precheck_coffee(request, pot):
    """
    Return a response if a precondition for coffee requests is not satisfied,
    else None.
    """
    if pot.is_teapot:
        return render(request, 'django_htcpcp_tea/418.html', status=418)

    if not pot.brew_coffee:
        return render(
            request,
            'django_htcpcp_tea/503.html',
            {'error_reason': 'Pot out of service. No coffee or tea available.'},
            status=503,
        )

    return None


def _precheck_teapot(request, pot, tea):
    """
    Return a response if a precondition for tea requests is not satisfied, else
    None.
    """
    if request.htcpcp_message_type == 'start':
        if not tea:  # Require tea type only when starting a new beverage
            alternatives = build_alternates(index_pot=pot)
            context = {'alternatives': alternatives}
            response = render(request, 'django_htcpcp_tea/options.html', context, status=300)
            response.htcpcp_alternates = alternatives
            return response
        elif tea not in pot.supported_teas.values_list('name'):
            return render(
                request,
                'django_htcpcp_tea/503.html',
                {'error_reason': '{} is not available for this pot'.format(tea.capitalize)},
                status=503,
            )
    return None


def _finalize_beverage(request, pot, beverage_name, additions):
    """
    Return a response to the beverage request according to the HTCPCP standard
    without referencing the user's session.
    """
    context = {
        'pot': pot,
        'beverage': beverage_name,
        'additions': additions,
    }

    if request.htcpcp_message_type == 'start':
        if beverage_name == 'coffee':
            # Display alternatives when brewing coffee per RFC 7168 section 2.1.1
            context['alternatives'] = build_alternates(index_pot=pot)
        response = render(request, 'django_htcpcp_tea/brewing.html', context, status=202)  # Accepted
    else:  # request.htcpcp_message_type == 'stop':
        if any(addition.is_milk for addition in additions):
            response = render(request, 'django_htcpcp_tea/pouring.html', context, status=200)  # Ok
        else:
            response = render(request, 'django_htcpcp_tea/finished.html', context, status=201)  # Created

    return response


def _finalize_beverage_with_session(request, pot, beverage_name, additions):
    """
    Return a response to the beverage request according to the HTCPCP standard
    by referencing the current state of the user's session.
    """
    session_key = 'htcpcp_pot_{}'.format(pot.id)
    pot_status = request.session.get(session_key)

    context = {
        'pot': pot,
        'beverage': beverage_name,
        'additions': additions,
    }

    # TODO add brew time and additions display to finished template

    if pot_status:
        # 'stop' requests may not be able to reproduce the name of a beverage
        # or the additions that were requested, so override these values in the
        # context with the ones stored in the user's session.
        context['beverage'] = pot_status['beverage']
        context['additions'] = pot_status['additions']

        if request.htcpcp_message_type == 'start':
            context['error_reason'] = 'Pot is busy and cannot start a new beverage.'
            response = render(request, 'django_htcpcp_tea/503.html', context, status=503)
        else:  # htcpcp_message_type == 'stop'
            if request.method == 'WHEN':
                if pot_status['currently_pouring']:
                    response = render(request, 'django_htcpcp_tea/finished.html', context, status=201)
                    del request.session[session_key]
                else:
                    context['error_reason'] = 'No milk is being poured. Please stop shouting "WHEN!"'
                    return render(request, 'django_htcpcp_tea/400.html', context, status=400)
            else:
                if pot_status['currently_pouring']:
                    context['error_reason'] = 'Milk is currently being poured. Please say "WHEN"'
                    response = render(
                        request, 'django_htcpcp_tea/400.html', context, status=400)
                elif pot_status['needs_milk']:  # Stop brewing and begin pouring milk
                    response = render(request, 'django_htcpcp_tea/pouring.html', context, status=200)
                    request.session[session_key] = {
                        **pot_status,
                        'needs_milk': False,
                        'currently_pouring': True,
                    }
                else:  # Stop brewing. No milk required.
                    response = render(request, 'django_htcpcp_tea/finished.html', context, status=201)
                    del request.session[session_key]
    elif request.htcpcp_message_type == 'start':
        # New session, and the client requested a new beverage
        if beverage_name == 'coffee':
            # Display alternatives when brewing coffee per RFC 7168 section 2.1.1
            context['alternatives'] = build_alternates(index_pot=pot)
        response = render(request, 'django_htcpcp_tea/brewing.html', context, status=202)  # Accepted
        request.session[session_key] = {
            'beverage': beverage_name,
            # Serialize the requested additions to as dictionaries for storage
            # in the user's session since we do not need actual Addition objects
            # to display the additions during future requests.
            'additions': [{'name': a.name, 'get_type_display': a.get_type_display()} for a in additions],
            'needs_milk': any(addition.is_milk for addition in additions),
            'currently_pouring': False,
            'start_time': datetime.utcnow().timestamp()
        }
    else:
        reason = ("No beverage is being brewed by this pot, but the "
                  "request did not indicate that a new beverage should be "
                  "brewed")
        response = render(request, 'django_htcpcp_tea/400.html', {'error_reason': reason}, status=400)

    return response


if htcpcp_settings.DISABLE_CSRF:
    # Mark the HTCPCP view function as being exempt from the CSRF view
    # protection. This is the same as using the csrf_exempt decorator
    # from django.views.decorators.csrf.
    brew_pot.csrf_exempt = True



#!/usr/bin/env python3
#  Copyright (c) 2019 Brian Schubert
#
#  This file is distributed under the MIT License. If a copy of the
#  MIT License was not distributed with this file, you can obtain one
#  at https://opensource.org/licenses/MIT.

import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Web Environment',
    'Framework :: Django',
    'Framework :: Django :: 2.0',
    'Framework :: Django :: 2.1',
    'Framework :: Django :: 2.2',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    "Topic :: Internet :: WWW/HTTP",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name='django-htcpcp-tea',
    version='0.2.1',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    license='MIT',
    author='Brian Schubert',
    url='https://github.com/blueschu/django-htcpcp-tea',
    description="Django app implementing HTCPCP-TEA as defined in RFC 7168.",
    long_description=README,
    long_description_content_type='text/x-rst',
    keywords="htcpcp django rfc-2324 rfc-7168 coffee tea",
    classifiers=CLASSIFIERS,
    install_requires=['django>=2.0'],
    python_requires='>=3',
    test_suite='runtests.run_tests',
)



# Generated by Django 2.2.2 on 2019-06-20 12:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_htcpcp_tea', '0003_auto_20190620_1205'),
    ]

    operations = [
        migrations.AddField(
            model_name='addition',
            name='type',
            field=models.CharField(choices=[('MLK', 'Milk'), ('SYP', 'Syrup'), ('SWT', 'Sweetener'), ('SPC', 'Spice'), ('ACL', 'Alcohol'), ('SUG', 'Sugar'), ('OTR', 'Other')], default='OTR', max_length=3, verbose_name='Addition type'),
            preserve_default=False,
        ),
    ]
